import numpy as np
import matplotlib.pyplot as plt
import scipy.io.wavfile as wavf
import speech_recognition as sr
from sklearn.metrics import mean_squared_error 
from scipy.fft import fft, ifft, fftfreq

class MicrophoneArray:
    # speed of sound 
    c = 346

    def __init__(self, num_mic=8, l_total=0.14):
        # Define Microphone Array

        self.num_mic = num_mic
        
        # define the spacing between microphones
        self.d = l_total/self.num_mic


        # initialize sample rate and signal size
        self.N = None
        self.sample_rate = None

        self.N_voice = 0 
        self.N_music = 0 
        self.target_fs = [75, 125, 250, 500, 1000, 2000, 3000, 4000, 5000, 8000, 10000, 20000]

        self.sources_v = []
        self.sources_m = []

    # phase factor for a fourier space signal, angle defined as difference from forward 
    # specific to frequency bin 
    def array_factor(self, angle, fs): 
        time_delays = np.arange(self.num_mic) * self.d * np.sin(angle) / self.c
        af= np.exp(np.multiply((-2j * np.pi * time_delays).reshape(-1,1), abs(np.array(fs)).reshape(1,-1)))
        return np.atleast_2d(af)

    def array_factor_(self, angle, f):
        return np.exp(-2j * np.pi * self.d * (f /self.c) * np.arange(self.num_mic) * np.sin(angle)).reshape(-1,1)


    # AUDIO MANIPULATION
    @staticmethod
    def read(f):
        fs, sound  = wavf.read(f)
        return sound, fs
    
    @staticmethod
    def write(data, fs_, name='audio_output/out.wav'):
        wavf.write(name, fs_, data)

    @staticmethod
    def psnr(sound, original):
        mse = mean_squared_error(sound, original)
        return 10*np.log(np.max(original)**2/mse)
    
    @staticmethod
    def snr(sound, original):
        # Calculate the power of the clean and noisy signals
        signal_power = np.sum(original**2) / len(original)
        noise_power = np.sum((original - sound)**2) / len(original)

        # Calculate SNR in decibels
        return 10 * np.log10(signal_power / noise_power)
    
    # compare audio to a list of signals
    def compare_all(self, sound, originals):
        psnrs = []
        for voice in originals:
            psnrs.append(self.psnr(sound[:self.N], voice[:self.N]))

        psnrs = np.array(psnrs)
        return psnrs
    
    # read in audio
    def load_audio(self, f, type ="v"):
        sound, fs = self.read(f)
        sample_length = len(sound)
        
        # update class sample rate and number of samples to current sound clip 
        if self.sample_rate is None: self.sample_rate = fs
        
        if self.N is None: self.N = sample_length
        elif self.N > sample_length:
            self.N = sample_length

        if type == "v":
            self.N_voice += 1
            sid=self.N_voice
            self.sources_v.append({'signal':sound, 'sid':sid, 'name':f'{type}{abs(sid)}'})


        if type == "m":
            self.N_music += 1
            sid=self.N_music
            self.sources_m.append({'signal':sound, 'sid':sid, 'name':f'{type}{abs(sid)}'})

        
        return sound

    def load_write(self, audio, name='audio_output/out.wav'):
        self.write(audio.squeeze().astype(np.int16), self.sample_rate, name)


    # randomly select audio based ont he number of voice sources and numbcer of music soures desired
    # returns voice sources then music sources 
    def randomize_experiment(self, Nv, Nm):

        v_in = np.random.choice(np.arange(0,self.N_voice), Nv, replace = False)
        m_in = np.random.choice(np.arange(0,self.N_music), Nm, replace = False)

        sounds = []
        for i in v_in:
            sounds.append(self.sources_v[i]['signal'][:self.N])

        for i in m_in:
            sounds.append(self.sources_m[i]['signal'][:self.N])
        
        return sounds
    
    # select audio based on names
    def get_sounds(self, names):
        sound_list = []
        # to make sure the user defined order is maintained
        for name in names:
            sound_list.append([d['signal'][:self.N] for d in (self.sources_v + self.sources_m) if d['name']== name][0])
        return sound_list
        
    # Anechoic model (no attenuation) 
    # delay signals in the time domain 
    def microphone_delays(self, signal, angle):
        # add noise in the future 
        additive_noise = np.zeros((self.num_mic, self.N))
        out = np.zeros((self.num_mic, self.N))

        # this method causes a signal power reduction
        delays_shift = self.sample_rate * (np.arange(self.num_mic) * self.d * np.sin(angle)/self.c)
        for i, delay in enumerate(delays_shift):
            out[i] = np.roll(signal, int(delay))
            
        return out + additive_noise
    
    # generates phasor tone 
    def generate_tones(self, frequencies):
        t = np.arange(self.N)/self.sample_rate # time vector
        signals = []
        for f in frequencies:
            signals.append(np.exp(2j * np.pi * f * t))

        return np.real(signals)
    
    # adds which noise to signal to provide a specific SNR
    def add_white_noise(self, signal, SNR=30):
        RMS_signal_sq= np.mean(signal**2)
        RMS_noise = np.sqrt(RMS_signal_sq/(10**(SNR/10)))
        noise = 1j * np.random.normal(0, RMS_noise, signal.shape).astype('complex128') +  np.random.normal(0, RMS_noise, signal.shape).astype('complex128')
        return signal.astype('complex128') + noise
    
        # scale signal to a desired power 
    def scale_power(self, original, target):
        # Calculate power of the original signal
        power_original = np.mean(original**2)

        # Calculate power of the target signal
        power_target = np.mean(target**2)

        # Calculate the scaling factor
        scaling_factor = np.sqrt(power_original / power_target)

        return target * scaling_factor

    # combine sounds and return frequency domain signal of each mircophone
    def combine_freq_domain(self, incoming_angles, signals):
        fs = fftfreq(self.N, d=1/self.sample_rate)
        signals = fft(signals, axis=1)
        
        combined=np.zeros((self.num_mic, self.N), dtype = 'complex128')
        for sig, angle in zip(signals, incoming_angles):
            direction = self.array_factor(angle, fs)

            combined += np.multiply(direction, sig)

        # combined= (ifft(combined))
        return combined
    
    # generate frequency mask
    def generate_masks(self):
        fs = abs(fftfreq(self.N, d=1/self.sample_rate))
        fs = np.repeat(fs[np.newaxis, :], self.num_mic, axis=0)
        logscale = np.log(self.target_fs)

        bins = [(fs <= np.exp((logscale[0] + logscale[1])/2))]
        for i in np.arange(len(logscale) - 2):
            low = np.exp((logscale[i] + logscale[i+1])/2)
            high = np.exp((logscale[i+1] + logscale[i+2])/2)
            bins.append(((fs <= high) & (fs > low)))

        bins.append((fs >= np.exp((logscale[-1] + logscale[-2])/2)))

        return bins


    # BEAMFORMING METHODS
    # Conventional, delay-and-sum, beamformer applied to each freq
    # outputs frequency domain weights
    def w_ds(self, steering_ang, f_, sig=None, custom_scaling=None):
        w = self.array_factor(steering_ang, f_)

        if custom_scaling is not None: 
            # example scalings:
            # scaling = np.linspace(-1, 1, self.num_mic)
            # scaling = np.power(scaling, 2) + 2
            w = w * custom_scaling

        # w /= np.sum(np.abs(w), axis=0)
        w /= w.conj().T @ w

        return w

    # Maximum Variance Distortionless Response Beamformer
    # Capon/MVDR beamformer applied to each freq
    # outputs frequency domain weights
    def w_mvdr(self, steering_ang, f_, sig):

        steering_vect = self.array_factor(steering_ang, f_)  # steering vector in the desired direction theta
        
        R = sig @ sig.conj().T # Calc covariance matrix. gives a Nr x Nr covariance matrix of the samples
        Rinv = np.linalg.pinv(R) # pseudo inverse
        
        w = (Rinv @ steering_vect)/(steering_vect.conj().T @ Rinv @ steering_vect)  # MVDR/Capon equation! numerator is 3x3 * 3x1, denominator is 1x3 * 3x3 * 3x1, resulting in a 3x1 weights vector
        return w
    

    # process frequency signal imput with desired beamforming method for a given steering angle
    def listen(self, steering_angle, input_f, method ="windowed", w= None, model = w_mvdr):
        fs = fftfreq(self.N, d=1/self.sample_rate)
        output = np.zeros(self.N, dtype = 'complex128')

        if method == "windowed":
            bins = self.generate_masks()
            for bin, freq_bin in zip(bins, self.target_fs):
                input_binned = np.where(bin, input_f, 0)
                output += self.listen_bin(steering_angle, input_binned, freq_bin, w=w, model = model)

        else:
            positive_fs = fs[:int(self.N/2-1)] if self.N % 2 == 0 else fs[:int((self.N-1)/2)]
            for f in positive_fs:
                output += self.listen_bin(steering_angle, input_f, f, w=w, model = model)
            
        
        # output = self.scale_power(np.mean(input, axis=0), output)
        return np.real(ifft(output))
    
    # apply beamforming to a single frequency bin 
    def listen_bin(self, steering_angle, input, bin_f, w= None, model = w_mvdr):
        # input = fft(input, axis=1)

        # if weights are no provided 
        if w is None:
            w = model(self, steering_angle, bin_f, input)

        # apply weights 
        r_weighted = w.conj().T @ input 
        # r_weighted = np.multiply(w.conj(), input)

        # sums all frequencies if individual frequencies weighted seperately 
        r_weighted = np.sum(r_weighted, axis=0)

        r_weighted = r_weighted.squeeze()
        
        return r_weighted
    
    # perform beamforming for all possible steering angles 
    def listen_pattern_fixed_input(self, input, originals, labels, method ="windowed", model=w_mvdr):
        theta_scan = np .linspace(-np.pi/2, np.pi/2, 180)
        psnr = []
        for theta_i in theta_scan:
            directed_listen = np.real(self.listen(theta_i, input, method = method ,model=model).T)
            # compare to each sound file stored 
            psnr_dir = self.compare_all(directed_listen, originals)
            psnr.append(psnr_dir)
        
        fig, ax = plt.subplots()
        ax.plot(theta_scan* 180 / np.pi, psnr)
        ax.axvline(x=0, linestyle='--', color='k')
        ax.set_title("PSNR of Each Source versus Steering Angle Direction")
        ax.legend(labels + ["Head Steering PSNR"])
        ax.set_ylabel("PSNR")
        ax.set_xlabel("Steering Angle")
        ax.set_ylim([-10, 75])
        plt.show()

    # plot the percent improvement in PSNR for fine-tuning to true incoming angle
    def listen_improvement(self, input, originals, angles, method ="windowed", model=w_mvdr):
        psnr = []

        # always evalauate at forward direction regardless of a sounce being present
        directed_listen = np.real(self.listen(0, input, method = method ,model=model).T)
        psnr_dir = self.compare_all(directed_listen, originals)
        psnr.append(psnr_dir)

        for theta_i in angles:
            # listen in the target direction
            directed_listen = np.real(self.listen(theta_i, input, method = method ,model=model).T)
            
            # compare to each sound file stored 
            psnr_dir = self.compare_all(directed_listen, originals)
            psnr.append(psnr_dir)
        psnr = np.array(psnr)

        metrics = np.zeros((2, len(originals)))
        
        for i in np.arange(len(originals)):
            # percent improvement in psnr versus headtracking
            metrics[0,i] = (np.max(psnr[:,i])- psnr[0,i])/psnr[0,i]
    
            #percent reduction in psnr versus headtracking
            metrics[1,i] = (psnr[0,i] - np.min(psnr[:,i]))/psnr[0,i]
        return metrics

    # BEAMPATTERN
    def beam_pattern(self, steering_angle, input_f, target_f, model = w_mvdr):
        
        # calcualte weights based on fourier domain input singals 
        # input = fft(input, axis=1)
        w = model(self, steering_angle, target_f, sig = input_f)

        # create tone representing the observing (plotted frequency) 
        t = np.arange(input_f.shape[1])/self.sample_rate # time vector
        tone = self.generate_tones([target_f])
        tone = fft(tone)

        # apply weighting to single tone at target plotting frequency 
        theta_scan = np .linspace(-np.pi/2, np.pi/2, 10000)
        dB = []
        for theta_i in theta_scan:
            
            # array factor (as if steering angle was this direction) 
            # a_scan = self.array_factor(theta_i, target_f) # array factor
            a_scan = self.array_factor(theta_i, target_f)
            r_scan = a_scan @ tone
            
            # apply weights 
            r_weighted = w.conj().T @ r_scan 

            r_weighted = ifft(r_weighted)
            dB.append(10*np.log10(np.var(r_weighted))) # power in signal, in dB so its easier to see small and large lobes at the same time
        dB -= np.max(dB) # normalize
        
        return dB, theta_scan
