#%%
import numpy as np
import matplotlib.pyplot as plt
from scipy.signal import argrelextrema
from matplotlib.cm import ScalarMappable
from matplotlib.colors import Normalize, LinearSegmentedColormap, to_rgba

#%%

def delay_and_sum (N, d_total, theta, f, phi):
    # assumptions: 
    #           - noise signals at microphones have the same energy, 
    #           - attentuation factors equal to one
    #           - source in the field

    # sound velocity in air (m/s)
    c = 346
    d = d_total/N
    

    a = (np.sin(N*np.pi*d*np.outer(f, (np.cos(phi)-np.cos(theta)))/c))
    b = (N*np.sin(np.pi*d*np.outer(f, (np.cos(phi)-np.cos(theta)))/c))
    beam_pattern = np.abs(np.divide(a, b, out=np.ones_like(a), where=b!=0))
    
    # convert to decibels 
    beam_pattern_db = 10 * np.log10(beam_pattern)

    return np.squeeze(beam_pattern_db)


def w_mvdr(theta, r, d_total, N):
    d = d_total/N
    a = np.exp(-2j * np.pi * d * np.arange(Nr) * np.sin(theta)) # steering vector in the desired direction theta
    a = a.reshape(-1,1) # make into a column vector (size 3x1)
    R = r @ r.conj().T # Calc covariance matrix. gives a Nr x Nr covariance matrix of the samples
    Rinv = np.linalg.pinv(R) # 3x3. pseudo-inverse tends to work better/faster than a true inverse
    w = (Rinv @ a)/(a.conj().T @ Rinv @ a) # MVDR/Capon equation! numerator is 3x3 * 3x1, denominator is 1x3 * 3x3 * 3x1, resulting in a 3x1 weights vector

    theta_scan = np.linspace(-1*np.pi, np.pi, 1000) # 1000 different thetas between -180 and +180 degrees
    results = []
    for theta_i in theta_scan:
        w = w_mvdr(theta_i, r) # 3x1
        r_weighted = w.conj().T @ r # apply weights
        power_dB = 10*np.log10(np.var(r_weighted)) # power in signal, in dB so its easier to see small and large lobes at the same time
        results.append(power_dB)
        results -= np.max(results) # normalize
   
    return w


# find half power beam width 
# input one full pattern of one frequency and steering angle
def find_HPBW(pattern, power=-3): 
    mask = np.where(pattern > power)
    # find the first and last index of above -3dB (half power)
    half_pwr_idx = (np.min(mask), np.max(mask))
    
    # largest angle minus smallest angle above 3dB
    return phi[half_pwr_idx[1]], phi[half_pwr_idx[0]], mask

# Rayleigh criterion for resolution of two waves of equal strength and equal frequency
def find_Rayleigh(pattern): 
    # find the main lobe location 
    idx_peak = np.argmax(pattern)

    # find the zeros locations (minima)
    zeros_idx = argrelextrema(pattern, np.less, mode='wrap')[0]

    if len(zeros_idx) > 0:
        # determine the closest zero
        d_nearest_zero = np.min(abs(idx_peak - zeros_idx))
    else: 
        d_nearest_zero = np.min([abs(idx_peak - 0), abs(idx_peak - (len(pattern)-1))])

    # return index spacing in equivalent degrees (phi spaced evenly)
    return np.rad2deg(phi[d_nearest_zero] - phi[0])

def find_max_mean_sidelobe(pattern): 
    # zeros_idx = argrelextrema(pattern, np.less)[0]
    lobes_idx = argrelextrema(pattern, np.greater, mode='wrap')[0]

    maximas = pattern[lobes_idx] 
    # sort in reverse order and grab everything except for the max
    side_lobes = np.sort(maximas)[::-1][1:]
    if len(side_lobes) > 0: 
        return side_lobes[0], np.mean(side_lobes)
    else: 
        return 0.0, 0.0

def polar_plot(phi, select_pattern, name=""): 
    # analyze pattern 
    side = find_max_mean_sidelobe(select_pattern)
    hpbw = find_HPBW(select_pattern)
    ray = find_Rayleigh(select_pattern)

    fig = plt.figure(figsize=(7,3))
    fig.suptitle(f"Beam Pattern {name}")

    ax1 = fig.add_subplot(121)
    ax1.plot(phi, select_pattern)
    ax1.set_ybound(-50, 0)
    ax1.set_ylabel("dB")
    ax1.set_xlabel("phi (rad)")
    ax1.axvspan(hpbw[0], hpbw[1], alpha=0.2, color='red')

    # polar plot
    ax2 = fig.add_subplot(122, polar=True)
    ax2.plot(phi, select_pattern)
    ax2.axvspan(hpbw[0], hpbw[1], alpha=0.2, color='red')

    ax2.set_thetamin(0)
    ax2.set_thetamax(180)
    ax2.set_xlabel("dB")
    ax2.set_ylabel("phi (rad)")
    ax2.set_ylim([-50, 0])

    plt.savefig(f'{name}.png', dpi=300)
    plt.show()
    

    print(f"The HPBW is {np.rad2deg(hpbw[0] - hpbw[1]):0.02f} deg")
    print(f"The Rayleigh Criterion distance is {ray:0.02f} deg")
    print(f"The max sidelobe power is {side[0]:0.02f} dB")
    print(f"The mean sidelobe power is {side[1]:0.02f} dB")

       
def plot_all_metrics(ds_pattern_db, f_array, steering_angle, phi):

    # evaluate the metrics for all frequencies and all possible steerign angles 
    ang_res = np.zeros((len(steering_angle), len(f_array), 4))
    for i in np.arange(len(steering_angle)):
        for j in np.arange(len(f_array)):
            pattern = ds_pattern_db[i, j]
            side = find_max_mean_sidelobe(pattern)
            hpbw_minmax, _ = find_HPBW(pattern)
            hpbw = np.rad2deg(hpbw_minmax[0] - hpbw_minmax[1])
            ray = find_Rayleigh(pattern)

            ang_res[i, j, 0] = ray
            ang_res[i, j, 1] = hpbw
            ang_res[i, j, 2] = side[0]
            ang_res[i, j, 3] = side[1]

    name = "Rayleigh Criteria"
    plt.imshow(np.clip(ang_res[:, :, 0].T, 0, 180), extent=(np.min(steering_angle), np.max(steering_angle), 
                                        np.min(f_array), np.max(f_array)), 
            aspect='auto', origin='lower', cmap='viridis')
    plt.xlabel("phi (rad)")
    plt.ylabel("Frequency (Hz)")
    plt.colorbar(label='Rayleigh Criterion (deg)')
    plt.title(name)
    plt.tight_layout()
    plt.savefig(f'{name}.png', dpi=300)
    plt.show()

    name = "Half Power Bandwidth"
    plt.imshow(ang_res[:, :, 1].T, extent=(np.min(steering_angle), np.max(steering_angle), 
                                        np.min(f_array), np.max(f_array)), 
            aspect='auto', origin='lower', cmap='viridis')
    plt.xlabel("phi (rad)")
    plt.ylabel("Frequency (Hz)")
    plt.colorbar(label='Half Power Bandwidth (deg)')
    plt.title(name)
    plt.tight_layout()
    plt.savefig(f'{name}.png', dpi=300)
    plt.show()

    name = "Max Sidelobe Level"
    plt.imshow(ang_res[:, :, 2].T, extent=(np.min(steering_angle), np.max(steering_angle), 
                                        np.min(f_array), np.max(f_array)), 
            aspect='auto', origin='lower', cmap='viridis')
    plt.xlabel("phi (rad)")
    plt.ylabel("Frequency (Hz)")
    plt.colorbar(label='Max Sidelobe Level (dB)')
    plt.title(name)
    plt.tight_layout()
    plt.savefig(f'{name}.png', dpi=300)
    plt.show()

def plot_over_importance(steering_angle, select_pattern=None, name=''):
    # importance estimated from paper 
    f_array_few = [125, 250, 500, 1000, 2000, 4000, 8000]
    f_importance_est_few = [0.015, 0.04, 0.125, 0.20, 0.32, 0.25, 0.05]

    # estimate importance curve
    f_array= np.logspace(1, 3.91)
    f_importance_est_curve = np.poly1d(np.polyfit(f_array_few, f_importance_est_few, 3))
    f_importance_est = f_importance_est_curve(f_array)
    f_importance_est = np.abs(f_importance_est)/np.sum(f_importance_est)
    
    select_pattern = (delay_and_sum(num_mic, total_l, steering_angle, f_array, phi))
    select_pattern_few = (delay_and_sum(num_mic, total_l, steering_angle, f_array_few, phi))

    def find_HPBW_importance(patterns, f, f_importance):
        hpbw = np.zeros((len(f), 2))
        importance = np.zeros((len(f), len(phi)))
        for j in range(len(f)):
            pattern = patterns[j]
            packed_hpbw = find_HPBW(pattern)
            hpbw[j] = packed_hpbw[0:2]
            importance[j][packed_hpbw[2]] = f_importance[j]
            # side = find_max_mean_sidelobe(pattern)
            # hpbw_ang = np.rad2deg(hpbw_minmax[0] - hpbw_minmax[1])
            # ray = find_Rayleigh(pattern)
        return hpbw, importance
    

    # hpbw, imp = find_HPBW_importance(select_pattern, f_array, f_importance_est)
    hpbw_few, imp_few = find_HPBW_importance(select_pattern_few, f_array_few, f_importance_est_few)

    importance_phi = np.sum(imp_few, axis=0)
    hpbw_60 = find_HPBW(importance_phi, 1-0.6)
    print(f"The HPBW (60% of important speech frequencies) is {np.rad2deg(hpbw_60[0] - hpbw_60[1]):0.02f} deg")
    # plt.plot(phi, importance_phi)
    # plt.title("Percent HPBW over phi")
    # plt.show()

    fig = plt.figure(figsize=(6,8), constrained_layout=True)
    fig.suptitle(f"Beam Pattern & HPBW {name}")

    # to do only plot certain frequencies 
    ax1 = fig.add_subplot(211, polar=True)
    # for i in reversed(range(len(f_array))):
    #     ax1.axvspan(hpbw[i, 0], hpbw[i, 1], color='red', alpha = f_importance_est[i])

    # ax1.axvspan(hpbw_60[0], hpbw_60[1], color='orange', alpha = 0.5,
    #             label = f"60% HPBW: {np.rad2degs(hpbw_60[0] - hpbw_60[1]):0.02f} deg")
    
    for i in reversed(range(len(f_array_few))):
        ax1.plot(phi, select_pattern_few[i], linewidth = 2, label=f"{f_array_few[i]} Hz") #, linewidth = (f_importance_est[i]*5 +1))

    # ax1.vlines(hpbw_60[0], -30, 0, color='orange', linewidth= 2, label = f"60% HPBW: {np.rad2deg(hpbw_60[0] - hpbw_60[1]):0.02f} deg")
    # ax1.vlines(hpbw_60[1], -30, 0, color='orange', linewidth= 2)
    ax1.set_ybound(-30, 0)
    ax1.set_xlabel("dB")
    ax1.set_ylabel("phi (deg)")
    ax1.set_thetamin(0)
    ax1.set_thetamax(180)
    ax1.legend(ncol=3)

    # add text 

    # polar plots
    ax2 = fig.add_subplot(212)
    # for i in reversed(range(len(f_array))):
    #     ax2.axvspan(hpbw[i, 0], hpbw[i, 1], color='red', alpha = f_importance_est[i])

    for i in reversed(range(len(f_array_few))):
        ax2.axvspan(hpbw_few[i, 0], hpbw_few[i, 1], color='red', alpha = f_importance_est_few[i])
        ax2.plot(phi, select_pattern_few[i], color='k', linewidth = 2, 
                 alpha=f_importance_est_few[i]/np.max(f_importance_est_few))
                #  label = f"{f_array_few[i]} Hz") #, linewidth = (f_importance_est[i]*5 +1))

    ax2.vlines(hpbw_60[0], -30, 0, color='r', linewidth= 4, label = f"60% HPBW: {np.rad2deg(hpbw_60[0] - hpbw_60[1]):0.02f} deg")
    ax2.vlines(hpbw_60[1], -30, 0, color='r', linewidth= 4)
    ax2.set_ylabel("dB")
    ax2.set_xlabel("phi (rad)")
    ax2.set_ylim([-30, 0])
    # labelLines(ax2.get_lines(), align=False, yoffsets=-0.6, backgroundcolor="none")
    ax2.legend()

    cmap1 = LinearSegmentedColormap.from_list(None, [to_rgba('k', 0), 'k'])
    norm = Normalize(vmin=0, vmax=1)
    sm1 = ScalarMappable(cmap=cmap1, norm=norm)
    sm1.set_array([])

    cbar = plt.colorbar(sm1, ax=ax2, ticks=[0, 1], pad=0.02)
    cbar.ax.set_yticklabels(["low", "high"])
    cbar.ax.set_ylabel("Importance for Intelligibility", labelpad=-20)

    cmap2 = LinearSegmentedColormap.from_list(None, [to_rgba('r', 0), 'r'])
    sm2 = ScalarMappable(cmap=cmap2, norm=norm)
    sm2.set_array([])

    cbar = plt.colorbar(sm2, ax=ax2, ticks=[], pad=0.02)

    plt.savefig(f'{name}.png', dpi=300)
    plt.show()


    # move the colour bars to horizontal in the polar plot space 
    # label with percentage values and label frequerncy vs HPBW

def spacing_optimal(f):
    c = 346
    return c/(2*f)

f_sample = 16000
num_mic = 7
# total_l = spacing_optimal(1000) * num_mic
total_l = 0.14
phi = np.arange(0, np.pi, 1/f_sample) 


#%%
plot_over_importance(np.pi/2, name='90deg')
plot_over_importance(80*np.pi/180, name='80deg')
plot_over_importance(np.pi/4, name='45deg')
plot_over_importance(np.pi/8, name='22.5deg')
# plot_over_importance(np.pi/10)



#%%

#%%
ds_pattern_db_2000= delay_and_sum(num_mic, total_l, np.pi/2, 2000, phi)
polar_plot(phi, ds_pattern_db_2000, f"{num_mic} Elements, {total_l}m length, 2kHz")

# ds_pattern_db_3000= delay_and_sum(num_mic/2, total_l, np.pi/2, 4000, phi)
# polar_plot(phi, ds_pattern_db_3000, f"{num_mic} Elements, {total_l}m length, 4kHz")

# ds_pattern_db_1000= delay_and_sum(num_mic/2, total_l, np.pi/2, 1000, phi)
# polar_plot(phi, ds_pattern_db_1000, f"{num_mic} Elements, {total_l}m length, 1kHz")

 #%%
f_array= np.logspace(1, 4)
# f_array = np.linspace(100, 8000)
steering_angle = np.linspace(0, np.pi, 17)

ds_pattern_db = np.empty((len(steering_angle), len(f_array), len(phi)))
for i, ang in enumerate(steering_angle): 
    ds_pattern_db[i] = (delay_and_sum(8, 0.02, ang, f_array, phi))

#%%
# plot all frequencies
# select_pattern = ds_pattern_db[8]
# name=""

# fig = plt.figure(figsize=(12,5), constrained_layout=True)
# fig.suptitle(f"Beam Pattern {name}")

# f_color= np.array(f_array)
# # norm = Normalize(vmin=0, vmax=np.max(f_color))
# norm = LogNorm(vmin=np.min(f_color), vmax=np.max(f_color))
# cmap = plt.get_cmap('gist_heat').reversed()
# sm = ScalarMappable(cmap=cmap, norm=norm)
# sm.set_array([])
# colors = cmap(norm(f_color))

# ax1 = fig.add_subplot(121, polar=True)
# for i in (range(len(f_color))):
#     # ax1.plot(phi, select_pattern[i], color=colors[i], alpha = 0.4, linewidth=2) #, linewidth = (f_importance_est[i]*5 +1))
#     ax1.fill_between(phi, select_pattern[i], y2=-30, color=colors[i], alpha =0.2)
#     # ax1.axvspan(hpbw[i, 0], hpbw[i, 1], color='red', alpha = f_importance_est[i])
# ax1.set_ybound(-30, 0)
# ax1.set_xlabel("dB")
# ax1.set_ylabel("phi (deg)")
# ax1.set_thetamin(0)
# ax1.set_thetamax(180)

# # polar plot
# ax2 = fig.add_subplot(122)
# for i in reversed(range(len(f_color))):
#     ax2.plot(phi, select_pattern[i], color=colors[i], alpha = 0.4, linewidth=2) #, linewidth = (f_importance_est[i]*5 +1))
#     # ax2.axvspan(hpbw[i, 0], hpbw[i, 1], color='red', alpha = f_importance_est[i])
# ax2.set_ylabel("dB")
# ax2.set_xlabel("phi (rad)")
# ax2.set_ylim([-30, 0])

# cbar = plt.colorbar(sm, ax=ax2, label='Frequency', ticks=None)

# plt.savefig(f'{name}.png', dpi=300)
# plt.show()
        
    
# %%

# plotting the phase differences between the farthest spaces microphones
c = 346
f_array_few = np.array([125, 250, 500, 1000, 2000, 4000, 8000])
length_diff = total_l * np.cos(phi)

fig = plt.figure(figsize=(7,3))
ax1 = fig.add_subplot(121)
ax2 = fig.add_subplot(122)

phi_deg =phi * 180 / np.pi
fig.suptitle(f"Fraction of wavelength")

for f in f_array_few:
    wavelen = c/f
    frac = length_diff/wavelen
    
    ax1.plot(phi_deg, abs(frac), label = f'{f} Hz')
    ax2.plot(phi_deg, abs(frac))   
    ax1.set_ylabel("fraction of wavelength resolution")
    ax1.set_xlabel("phi")

ax2.set_ylim([0,1])
ax1.set_xlim([45, 3*45])
ax2.set_xlim([45, 3*45])
ax1.legend()

# %%
sample_freq = 44100